<?php

require('xstring.php');

class typewrapper
{
	protected $className;
	private $class;
	private $methods;
	static private $innerTypes = array('Int', 'String', 'Array', 'Bool', 'Float');

	protected function init()
	{
		// Inicializáció - lekérdezzünk az objektum metódusait
		$this->class = new ReflectionClass($this->className);
		$this->methods = $this->class->getMethods();
	}
	
	static private function parseMethodName($methodName)
	{
		// Egy metódus típusozott nevét felbontjuk alap névre és várt típusokra
		$types = array();
		$remainName = $methodName;
		while (mb_ereg('^.*_([A-Z][a-z]*)$', $remainName, $regs))
		{
			$types[] = $regs[1];
			$remainName = mb_substr($remainName, 0, mb_strlen($remainName) - mb_strlen($regs[1]) - 1);
		}
		return array($remainName, $types);
	}

	static private function isUsableAs($className, $baseClassName)
	{
		// Megállapítjuk, hogy adott osztályú objektum megfelel e az elvért osztálynak
		if (in_array($className, self::$innerTypes) || in_array($baseClassName, self::$innerTypes))
		{
			return $className == $baseClassName;
		}
		else
		{
			$class = new ReflectionClass($className);
			$neededClass = new ReflectionClass($baseClassName);
			return $class->getname() == $neededClass->getName() || $class->isSubclassOf($neededClass);
		}
	}

	private function findMethod($name, $types)
	{
		// Megtaláljuk (jobbik esetben) a típusoknak megfelelő metódust 
		$result = FALSE;
		$match = FALSE;
		foreach($this->methods as $index => $method)
		{
			if ($method->isPublic() && $method->getNumberOfRequiredParameters() == count($types))
			{
				list($methodName, $info) = self::parseMethodName($method->getName());
				if ($methodName == $name)
				{
					for($i = 0; $i < count($info); ++$i)
					{
						 if (!self::isUsableAs($types[$i], $info[$i])) break;
					}
					$match = $i == count($info);
				}
			}
			if ($match === TRUE)
			{
				$result = $index;
				break;
			}
		}
		if ($result === FALSE) throw new Exception("Method not found");
		return $result;
	}

	static private function getClassName($arg)
	{
		// Megállapítjuk egy objektum osztályát
		if (is_int($arg)) $result = 'Int';
		else if (is_array($arg)) $result = 'Array';
		else if (is_bool($arg)) $result = 'Bool';
		else if (is_float($arg)) $result = 'Float';
		else if (is_string($arg)) $result = 'String';
		else
		{
			foreach(get_declared_classes() as $class)
			{
				if (is_a($arg, $class))
				{
					$result = xstring::capitalize($class);
					break;
				}
			}
		}
		return $result;
	} 

	public function __call($name, $arguments)
	{
		$types = array();
		// Enchance argument with types
		foreach ($arguments as $index => $arg) $types[$index] = self::getClassName($arg); 
		try
		{
			$index = $this->findMethod($name, $types);
			return $this->methods[$index]->invokeArgs($this, $arguments);
		}
		catch (Exception $e) { print_r($e); }
	}
	
	public function __get($name)
	{
		try
		{
			$index = $this->findMethod("get{$name}", array());
			return $this->methods[$index]->invokeArgs($this, array());
		}
		catch (Exception $e) { print_r($e); }
	}
	
	public function __set($name, $value)
	{
		try
		{
			$index = $this->findMethod("set{$name}", array($value));
			return $this->methods[$index]->invokeArgs($this, array($value));
		}
		catch (Exception $e) { print_r($e); }
	}
}

?>